"""migrate records from old format to new.

Revision ID: 71b84ccc31bb
Revises: f2772c2c64a6
Create Date: 2018-02-14 12:04:34.688790

"""

# revision identifiers, used by Alembic.
revision = '71b84ccc31bb'
down_revision = 'f2772c2c64a6'

from alembic import op
import sqlalchemy as sa

import requests

from waiverdb.api_v1 import get_resultsdb_result
from waiverdb.models import Waiver


def convert_id_to_subject_and_testcase(result_id):
    try:
        result = get_resultsdb_result(result_id)
    except requests.HTTPError as e:
        if e.response.status_code == 404:
            raise RuntimeError('Result id %s not found in Resultsdb' % (result_id))
        else:
            raise RuntimeError('Failed looking up result in Resultsdb: %s' % e)
    except Exception as e:
        raise RuntimeError('Failed looking up result in Resultsdb: %s' % e)
    if 'original_spec_nvr' in result['data']:
        subject = {'original_spec_nvr': result['data']['original_spec_nvr'][0]}
    else:
        if result['data']['type'][0] == 'koji_build' or \
           result['data']['type'][0] == 'bodhi_update':
            SUBJECT_KEYS = ['item', 'type']
            subject = dict([(k, v[0]) for k, v in result['data'].items()
                            if k in SUBJECT_KEYS])
        else:
            raise RuntimeError('Unable to determine subject for result id %s' % (result_id))
    testcase = result['testcase']['name']
    return (subject, testcase)


def upgrade():
    # Get a session asociated with the alembic upgrade operation.
    connection = op.get_bind()
    Session = sa.orm.sessionmaker()
    session = Session(bind=connection)

    try:
        # querying resultsdb for the corresponding subject/testcase.
        waivers = session.query(Waiver).all()
        for waiver in waivers:
            subject, testcase = convert_id_to_subject_and_testcase(waiver.result_id)
            waiver.subject = subject
            waiver.testcase = testcase
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()


def downgrade():
    # It shouldn't be possible to downgrade this change.
    # Because the result_id field will not be populated with data anymore.
    # If the user tries to downgrade "result_id" should be not null once again
    # like in the old version of the schema, but the value is not longer available
    raise RuntimeError('Irreversible migration')



"""Change waiver.subject column type to JSON

Revision ID: ce8a1351ecdc
Revises: 1797bff52162
Create Date: 2018-02-28 14:17:15.545322

"""

# revision identifiers, used by Alembic.
revision = 'ce8a1351ecdc'
down_revision = '1797bff52162'

import json
from alembic import op
from sqlalchemy import Text, text
from sqlalchemy.dialects.postgresql import JSON
from waiverdb.models.base import json_serializer


def upgrade():
    # Re-serialize all subject values to ensure they match the new, consistent
    # serialization we are using.
    connection = op.get_bind()
    for row in connection.execute('SELECT id, subject FROM waiver'):
        fixed_subject = json_serializer(json.loads(row['subject']))
        connection.execute(text('UPDATE waiver SET subject = :subject WHERE id = :id'),
                           subject=fixed_subject, id=row['id'])

    op.drop_index('ix_waiver_subject')
    op.alter_column('waiver', 'subject', type_=JSON, postgresql_using='subject::json')
    op.create_index('ix_waiver_subject', 'waiver', [text('CAST(subject AS TEXT)')])


def downgrade():
    op.drop_index('ix_waiver_subject')
    op.alter_column('waiver', 'subject', type_=Text)
    op.create_index('ix_waiver_subject', 'waiver', ['subject'])



# SPDX-License-Identifier: GPL-2.0+

from waiverdb.models import Waiver


def create_waiver(session, subject, testcase, username, product_version, waived=True,
                  comment=None, proxied_by=None):
    waiver = Waiver(subject, testcase, username, product_version, waived, comment,
                    proxied_by)
    session.add(waiver)
    session.flush()
    return waiver



# -*- coding: utf-8 -*-
#
# WaiverDB documentation build configuration file, created by
# sphinx-quickstart on Wed Mar 15 17:21:27 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

# This will cause the Flask application to be created with development configs
os.environ['DEV'] = 'true'

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))
import waiverdb  # noqa: E402


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
needs_sphinx = '1.3'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinxcontrib.autohttp.flask',
    'sphinxcontrib.issuetracker',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'WaiverDB'
copyright = u'2017, Red Hat, Inc. and others'
author = u'Red Hat, Inc. and others'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '.'.join(waiverdb.__version__.split('.')[:2])
# The full version, including alpha/beta/rc tags.
release = waiverdb.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'WaiverDBdoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'WaiverDB.tex', u'WaiverDB Documentation',
     u'Red Hat, Inc. and others', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'waiverdb', u'WaiverDB Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'WaiverDB', u'WaiverDB Documentation',
     author, 'WaiverDB', 'One line description of project.',
     'Miscellaneous'),
]


# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'https://docs.python.org/2/': 'python-intersphinx.inv'}

# Configuration for issue tracker magic linking.
issuetracker = None
issuetracker_url = 'https://pagure.io/'
issuetracker_project = 'waiverdb'

# Until Pagure support is merged: https://github.com/ignatenkobrain/sphinxcontrib-issuetracker/pull/15
import requests
from sphinxcontrib.issuetracker import Issue
PAGURE_URL = '{0.url}/{0.project}/issue/{1}'
PAGURE_API_URL = '{0.url}/api/0/{0.project}/issue/{1}'
HEADERS = {
    'User-Agent': 'sphinxcontrib-issuetracker'
}
def get(app, url):
    try:
        response = requests.get(url, headers=HEADERS)
        if response.status_code == requests.codes.ok:
            return response
        elif response.status_code != requests.codes.not_found:
            msg = 'GET {0.url} failed with code {0.status_code}'
            app.warn(msg.format(response))
    except IOError as e:
        msg = 'GET {0} failed with error: {1}'
        app.warn(msg.format(url, e))
def lookup_pagure_issue(app, tracker_config, issue_id):
    if not tracker_config.url:
        raise ValueError('URL required, try: https://pagure.io/')
    issue_url = PAGURE_URL.format(tracker_config, issue_id)
    response = get(app, PAGURE_API_URL.format(tracker_config, issue_id))
    if response:
        title = response.json()['title']
        closed = response.json()['status'] != 'Open'
        return Issue(id=issue_id, title=title, closed=closed, url=issue_url)
    else:
        return Issue(id=issue_id, title=None, closed=False, url=issue_url)

def setup(app):
    app.connect('issuetracker-lookup-issue', lookup_pagure_issue)



# SPDX-License-Identifier: GPL-2.0+
"""
This module contains a set of `SQLAlchemy event`_ hooks.

To use these hooks, you must register them with SQLAlchemy
using the :func:`sqlalchemy.event.listen` function.

.. _SQLALchemy events:
    https://docs.sqlalchemy.org/en/latest/orm/events.html
"""
from __future__ import unicode_literals

import logging

from flask_restful import marshal
import fedmsg
import stomp
import json
from waiverdb.fields import waiver_fields
from waiverdb.models import Waiver
from waiverdb.utils import stomp_connection
from flask import current_app

_log = logging.getLogger(__name__)


def publish_new_waiver(session):
    """
    A post-commit event hook that emits messages to a message bus. The messages
    can be published by either fedmsg-hub with zmq or stomp.

    This event is designed to be registered with a session factory::

        >>> from sqlalchemy.event import listen
        >>> listen(MyScopedSession, 'after_commit', publish_new_waiver)

    The emitted message will look like::

        {
          "username": "jcline",
          "i": 4,
          "timestamp": 1489686124,
          "msg_id": "2017-80e46243-e6f5-46df-8dcd-4d17809eb298",
          "topic": "org.fedoraproject.dev.waiverdb.waiver.new",
          "msg": {
            "comment": "Because I said so",
            "username": "http://jcline.id.fedoraproject.org/",
            "waived": true,
            "timestamp": "2017-03-16T17:42:04.209638",
            "product_version": "Satellite 6.3",
            "subject": "{\"a.nice.example\": \"this-is-a-really-nice-example\"}",
            "testcase": "t.e.s.t.case",
            "proxied_by": null,
            "id": 15
          }
        }

    Args:
        session (sqlalchemy.orm.Session): The session that was committed to the
            database. This session is not active and cannot emit SQL.

    """
    _log.debug('The publish_new_waiver SQLAlchemy event has been activated.')
    if current_app.config['MESSAGE_PUBLISHER'] == 'stomp':
        with stomp_connection() as conn:
            stomp_configs = current_app.config.get('STOMP_CONFIGS')
            for row in session.identity_map.values():
                if isinstance(row, Waiver):
                    _log.debug('Publishing a message for %r', row)
                    msg = json.dumps(marshal(row, waiver_fields))
                    kwargs = dict(body=msg, headers={}, destination=stomp_configs['destination'])
                    if stomp.__version__[0] < 4:
                        kwargs['message'] = kwargs.pop('body')  # On EL7, different sig.
                    conn.send(**kwargs)
    else:
        for row in session.identity_map.values():
            if isinstance(row, Waiver):
                _log.debug('Publishing a message for %r', row)
                fedmsg.publish(topic='waiver.new', msg=marshal(row, waiver_fields))



# SPDX-License-Identifier: GPL-2.0+

import json
from .utils import create_waiver
import datetime
from requests import ConnectionError, HTTPError
from mock import patch, Mock
from waiverdb import __version__


@patch('waiverdb.auth.get_user', return_value=('foo', {}))
def test_create_waiver(mocked_get_user, client, session):
    data = {
        'subject': {'subject.test': 'subject'},
        'testcase': 'testcase1',
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 201
    assert res_data['username'] == 'foo'
    assert res_data['subject'] == {'subject.test': 'subject'}
    assert res_data['testcase'] == 'testcase1'
    assert res_data['product_version'] == 'fool-1'
    assert res_data['waived'] is True
    assert res_data['comment'] == 'it broke'


@patch('waiverdb.api_v1.get_resultsdb_result')
@patch('waiverdb.auth.get_user', return_value=('foo', {}))
def test_create_waiver_legacy(mocked_get_user, mocked_resultsdb, client, session):
    mocked_resultsdb.return_value = {
        'data': {
            'type': ['koji_build'],
            'item': ['somebuild'],
        },
        'testcase': {'name': 'sometest'}
    }

    data = {
        'result_id': 123,
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 201
    assert res_data['username'] == 'foo'
    assert res_data['subject'] == {'type': 'koji_build', 'item': 'somebuild'}
    assert res_data['testcase'] == 'sometest'
    assert res_data['product_version'] == 'fool-1'
    assert res_data['waived'] is True
    assert res_data['comment'] == 'it broke'


@patch('waiverdb.api_v1.get_resultsdb_result')
@patch('waiverdb.auth.get_user', return_value=('foo', {}))
def test_create_waiver_with_original_spec_nvr_subject(mocked_get_user, mocked_resultsdb, client,
                                                      session):
    mocked_resultsdb.return_value = {
        'data': {
            'original_spec_nvr': ['somedata'],
        },
        'testcase': {'name': 'sometest'}
    }

    data = {
        'result_id': 123,
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 201
    assert res_data['username'] == 'foo'
    assert res_data['subject'] == {'original_spec_nvr': 'somedata'}
    assert res_data['testcase'] == 'sometest'
    assert res_data['product_version'] == 'fool-1'
    assert res_data['waived'] is True
    assert res_data['comment'] == 'it broke'


@patch('waiverdb.api_v1.get_resultsdb_result', side_effect=HTTPError(response=Mock(status=404)))
@patch('waiverdb.auth.get_user', return_value=('foo', {}))
def test_create_waiver_with_unknown_result_id(mocked_get_user, mocked_resultsdb, client, session):
    data = {
        'result_id': 123,
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
    }
    mocked_resultsdb.return_value.status_code = 404
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert res_data['message'].startswith('Failed looking up result in Resultsdb:')


@patch('waiverdb.auth.get_user', return_value=('foo', {}))
def test_create_waiver_with_no_testcase(mocked_get_user, client):
    data = {
        'subject': {'foo': 'bar'},
        'waived': True,
        'product_version': 'the-best',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 400
    # XXX - when we ditch result_id and make subject and testcase required
    # again, this assertion will change back to the original.
#   assert 'Missing required parameter in the JSON body' in res_data['message']['testcase']
    assert 'Either result_id or subject/testcase are required arguments.' in res_data['message']


@patch('waiverdb.auth.get_user', return_value=('foo', {}))
def test_create_waiver_with_malformed_subject(mocked_get_user, client):
    data = {
        'subject': 'asd',
        'testcase': 'qqq',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 400
    assert 'Must be a valid dict' in res_data['message']['subject']


@patch('waiverdb.auth.get_user', return_value=('foo', {}))
def test_non_superuser_cannot_create_waiver_for_other_users(mocked_get_user, client):
    data = {
        'subject': {'subject.test': 'subject'},
        'testcase': 'testcase1',
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
        'username': 'bar',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 403
    assert 'user foo does not have the proxyuser ability' == res_data['message']


@patch('waiverdb.auth.get_user', return_value=('bodhi', {}))
def test_superuser_can_create_waiver_for_other_users(mocked_get_user, client, session):
    data = {
        'subject': {'subject.test': 'subject'},
        'testcase': 'testcase1',
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
        'username': 'bar',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 201
    # a waiver should be created for bar by bodhi
    assert res_data['username'] == 'bar'
    assert res_data['proxied_by'] == 'bodhi'


def test_get_waiver(client, session):
    # create a new waiver
    waiver = create_waiver(session, subject={'subject.test': 'subject'},
                           testcase='testcase1', username='foo',
                           product_version='foo-1', comment='bla bla bla')
    r = client.get('/api/v1.0/waivers/%s' % waiver.id)
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert res_data['username'] == waiver.username
    assert res_data['subject'] == waiver.subject
    assert res_data['testcase'] == waiver.testcase
    assert res_data['product_version'] == waiver.product_version
    assert res_data['waived'] is True
    assert res_data['comment'] == waiver.comment


def test_404_for_nonexistent_waiver(client, session):
    r = client.get('/api/v1.0/waivers/foo')
    assert r.status_code == 404
    res_data = json.loads(r.get_data(as_text=True))
    message = (
        'The requested URL was not found on the server.  If you entered the '
        'URL manually please check your spelling and try again.')
    assert res_data['message'] == message


@patch('waiverdb.api_v1.AboutResource.get', side_effect=ConnectionError)
def test_500(mocked, client, session):
    r = client.get('/api/v1.0/about')
    assert r.status_code == 500
    res_data = json.loads(r.get_data(as_text=True))
    assert res_data['message'] == ''


def test_get_waivers(client, session):
    for i in range(0, 10):
        create_waiver(session, subject={"subject%d" % i: "%d" % i},
                      testcase="case %d" % i, username='foo %d' % i,
                      product_version='foo-%d' % i, comment='bla bla bla')
    r = client.get('/api/v1.0/waivers/')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 10


def test_pagination_waivers(client, session):
    for i in range(0, 30):
        create_waiver(session, subject={"subject%d" % i: "%d" % i},
                      testcase="case %d" % i, username='foo %d' % i,
                      product_version='foo-%d' % i, comment='bla bla bla')
    r = client.get('/api/v1.0/waivers/?page=2')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 10
    assert '/waivers/?page=1' in res_data['prev']
    assert '/waivers/?page=3' in res_data['next']
    assert '/waivers/?page=1' in res_data['first']
    assert '/waivers/?page=3' in res_data['last']


def test_obsolete_waivers_are_excluded_by_default(client, session):
    create_waiver(session, subject={'subject.test': 'subject'},
                  testcase='testcase1', username='foo',
                  product_version='foo-1')
    new_waiver = create_waiver(session, subject={'subject.test': 'subject'},
                               testcase='testcase1', username='foo',
                               product_version='foo-1', waived=False)
    r = client.get('/api/v1.0/waivers/')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['id'] == new_waiver.id
    assert res_data['data'][0]['waived'] == new_waiver.waived


def test_get_obsolete_waivers(client, session):
    old_waiver = create_waiver(session, subject={'subject.test': 'subject'},
                               testcase='testcase1', username='foo',
                               product_version='foo-1')
    new_waiver = create_waiver(session, subject={'subject.test': 'subject'},
                               testcase='testcase1', username='foo',
                               product_version='foo-1', waived=False)
    r = client.get('/api/v1.0/waivers/?include_obsolete=1')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 2
    assert res_data['data'][0]['id'] == new_waiver.id
    assert res_data['data'][1]['id'] == old_waiver.id


def test_filtering_waivers_by_subject_and_testcase(client, session):
    create_waiver(session, subject={'subject.test1': 'subject1'},
                  testcase='testcase1', username='foo-1', product_version='foo-1')
    create_waiver(session, subject={'subject.test2': 'subject2'},
                  testcase='testcase2', username='foo-2', product_version='foo-1')

    param = json.dumps([{'subject': {'subject.test1': 'subject1'}, 'testcase': 'testcase1'}])
    r = client.get('/api/v1.0/waivers/?results=%s' % param)
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['subject'] == {'subject.test1': 'subject1'}
    assert res_data['data'][0]['testcase'] == 'testcase1'


def test_filtering_waivers_by_subject_and_testcase_with_other_json_order(client, session):
    create_waiver(session, subject={'subject.test1': 'subject1'},
                  testcase='testcase1', username='foo-1', product_version='foo-1')

    param = json.dumps([{'testcase': 'testcase1', 'subject': {'subject.test1': 'subject1'}}])
    r = client.get('/api/v1.0/waivers/?results=%s' % param)
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['subject'] == {'subject.test1': 'subject1'}
    assert res_data['data'][0]['testcase'] == 'testcase1'


def test_filtering_waivers_by_multiple_results_subjects_and_testcases(client, session):
    create_waiver(session, subject={'subject.test1': 'subject1'},
                  testcase='testcase1', username='foo-1', product_version='foo-1')
    create_waiver(session, subject={'subject.test2': 'subject2'},
                  testcase='testcase2', username='foo-2', product_version='foo-1')
    create_waiver(session, subject={'subject.test3': 'subject3'},
                  testcase='testcase3', username='foo-2', product_version='foo-1')
    param = json.dumps([{'subject': {'subject.test1': 'subject1'}, 'testcase': 'testcase1'},
                        {'subject': {'subject.test2': 'subject2'}, 'testcase': 'testcase2'}])
    r = client.get('/api/v1.0/waivers/?results=%s' % param)
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 2
    assert res_data['data'][0]['subject'] == {'subject.test2': 'subject2'}
    assert res_data['data'][0]['testcase'] == 'testcase2'
    assert res_data['data'][1]['subject'] == {'subject.test1': 'subject1'}
    assert res_data['data'][1]['testcase'] == 'testcase1'


def test_filtering_waivers_by_subject_without_testcase(client, session):
    create_waiver(session, subject={'subject.test1': 'subject1'},
                  testcase='testcase1', username='foo-1', product_version='foo-1')
    create_waiver(session, subject={'subject.test2': 'subject2'},
                  testcase='testcase2', username='foo-2', product_version='foo-1')

    param = json.dumps([{'subject': {'subject.test1': 'subject1'}}])
    r = client.get('/api/v1.0/waivers/?results=%s' % param)
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['subject'] == {'subject.test1': 'subject1'}
    assert res_data['data'][0]['testcase'] == 'testcase1'


def test_filtering_waivers_by_product_version(client, session):
    create_waiver(session, subject={'subject.test1': 'subject1'},
                  testcase='testcase1', username='foo-1', product_version='release-1')
    create_waiver(session, subject={'subject.test2': 'subject2'},
                  testcase='testcase2', username='foo-1', product_version='release-2')
    r = client.get('/api/v1.0/waivers/?product_version=release-1')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['product_version'] == 'release-1'


def test_filtering_waivers_by_username(client, session):
    create_waiver(session, subject={'subject.test1': 'subject1'},
                  testcase='testcase1', username='foo', product_version='foo-1')
    create_waiver(session, subject={'subject.test2': 'subject2'},
                  testcase='testcase2', username='bar', product_version='foo-2')
    r = client.get('/api/v1.0/waivers/?username=foo')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['username'] == 'foo'


def test_filtering_waivers_by_since(client, session):
    before1 = (datetime.datetime.utcnow() - datetime.timedelta(seconds=100)).isoformat()
    before2 = (datetime.datetime.utcnow() - datetime.timedelta(seconds=99)).isoformat()
    after = (datetime.datetime.utcnow() + datetime.timedelta(seconds=100)).isoformat()
    create_waiver(session, subject={'subject.test1': 'subject1'},
                  testcase='testcase1', username='foo', product_version='foo-1')
    r = client.get('/api/v1.0/waivers/?since=%s' % before1)
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['subject'] == {'subject.test1': 'subject1'}
    assert res_data['data'][0]['testcase'] == 'testcase1'

    r = client.get('/api/v1.0/waivers/?since=%s,%s' % (before1, after))
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['subject'] == {'subject.test1': 'subject1'}
    assert res_data['data'][0]['testcase'] == 'testcase1'

    r = client.get('/api/v1.0/waivers/?since=%s' % (after))
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 0

    r = client.get('/api/v1.0/waivers/?since=%s,%s' % (before1, before2))
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 0


def test_filtering_waivers_by_malformed_since(client, session):
    create_waiver(session, subject={'subject.test1': 'subject1'},
                  testcase='testcase1', username='foo', product_version='foo-1')
    wrong_since = 123
    r = client.get('/api/v1.0/waivers/?since=%s' % wrong_since)
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 400
    assert res_data['message'] == "'since' parameter not in ISO8601 format"


def test_filtering_waivers_by_proxied_by(client, session):
    create_waiver(session, subject={'subject.test1': 'subject1'},
                  testcase='testcase1', username='foo-1', product_version='foo-1',
                  proxied_by='bodhi')
    create_waiver(session, subject={'subject.test2': 'subject2'},
                  testcase='testcase2', username='foo-2', product_version='foo-1')
    r = client.get('/api/v1.0/waivers/?proxied_by=bodhi')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['subject'] == {'subject.test1': 'subject1'}
    assert res_data['data'][0]['testcase'] == 'testcase1'


def test_jsonp(client, session):
    waiver = create_waiver(session, subject={'subject.test1': 'subject1'},
                           testcase='testcase1', username='foo', product_version='foo-1')
    r = client.get('/api/v1.0/waivers/%s?callback=jsonpcallback' % waiver.id)
    assert r.mimetype == 'application/javascript'
    assert 'jsonpcallback' in r.get_data(as_text=True)


def test_healthcheck(client):
    r = client.get('healthcheck')
    assert r.status_code == 200
    assert r.get_data(as_text=True) == 'Health check OK'


def test_get_waivers_with_post_request(client, session):
    """
    This tests that users can get waivers by sending a POST request with a long
    list of result subject/testcase.
    """
    results = []
    for i in range(1, 51):
        results.append({'subject': {'subject%d' % i: '%d' % i},
                        'testcase': 'case %d' % i})
        create_waiver(session, subject={"subject%d" % i: "%d" % i},
                      testcase="case %d" % i, username='foo %d' % i,
                      product_version='foo-%d' % i, comment='bla bla bla')
    data = {
        'results': results
    }
    r = client.post('/api/v1.0/waivers/+by-subjects-and-testcases', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 50
    subjects = []
    testcases = []
    for i in reversed(range(1, 51)):
        subjects.append({'subject%d' % i: '%d' % i})
        testcases.append('case %d' % i)
    assert [w['subject'] for w in res_data['data']] == subjects
    assert set([w['testcase'] for w in res_data['data']]) == set(testcases)
    assert all(w['username'].startswith('foo') for w in res_data['data'])
    assert all(w['product_version'].startswith('foo-') for w in res_data['data'])


def test_get_waivers_with_post_malformed_since(client, session):
    create_waiver(session, subject={'subject.test1': 'subject1'},
                  testcase='testcase1', username='foo', product_version='foo-1')
    data = {
        'since': 123
    }
    r = client.post('/api/v1.0/waivers/+by-subjects-and-testcases', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 400
    assert res_data['message'] == "'since' parameter not in ISO8601 format"


def test_about_endpoint(client):
    r = client.get('/api/v1.0/about')
    output = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert output['version'] == __version__
    assert output['auth_method'] == client.application.config['AUTH_METHOD']



# SPDX-License-Identifier: GPL-2.0+

from copy import copy
import pytest
from sqlalchemy import create_engine
from waiverdb.app import create_app


@pytest.fixture(scope='session')
def app(request):
    app = create_app('waiverdb.config.TestingConfig')
    # Establish an application context before running the tests.
    ctx = app.app_context()
    ctx.push()

    def teardown():
        ctx.pop()

    request.addfinalizer(teardown)
    return app


@pytest.fixture(scope='session')
def db(app):
    """Session-wide test database."""
    from waiverdb.models import db
    dbname = db.engine.url.database
    # In order to drop and re-create the database, we have to connect to
    # template1 database in special AUTOCOMMIT isolation level.
    dburl = copy(db.engine.url)
    dburl.database = 'template1'
    with create_engine(dburl).connect() as connection:
        connection.execution_options(isolation_level='AUTOCOMMIT')
        connection.execute('DROP DATABASE IF EXISTS {}'.format(dbname))
        connection.execute('CREATE DATABASE {}'.format(dbname))
    db.create_all()
    return db


@pytest.yield_fixture
def session(db, monkeypatch):
    """Patch Flask-SQLAlchemy to use a specific connection"""
    connection = db.engine.connect()
    transaction = connection.begin()

    # Patch Flask-SQLAlchemy to use our connection
    monkeypatch.setattr(db, 'get_engine', lambda *args: connection)

    yield db.session

    db.session.remove()
    transaction.rollback()
    connection.close()


@pytest.yield_fixture
def client(app):
    """A Flask test client. An instance of :class:`flask.testing.TestClient`
    by default.
    """
    with app.test_client() as client:
        yield client


@pytest.fixture()
def enable_kerberos(app, monkeypatch):
    monkeypatch.setitem(app.config, 'AUTH_METHOD', 'Kerberos')


@pytest.fixture()
def enable_ssl(app, monkeypatch):
    monkeypatch.setitem(app.config, 'AUTH_METHOD', 'SSL')



# SPDX-License-Identifier: GPL-2.0+
"""
CLI for creating new waivers against test results.
"""
import click
import requests
import json
import configparser

requests_session = requests.Session()


def validate_config(config):
    """
    Validates the configuration needed for WaiverDB
    :return: None or ClickException
    """
    # Verify that all necessary config options are set
    config_error = ('The config option "{0}" is required')
    if not config.has_option('waiverdb', 'auth_method'):
        raise click.ClickException(config_error.format('auth_method'))
    auth_method = config.get('waiverdb', 'auth_method')
    if auth_method not in ['OIDC', 'Kerberos', 'dummy']:
        raise click.ClickException('The WaiverDB authentication mechanism of '
                                   '"{0}" is not supported'.format(auth_method))
    required_configs = ['api_url']
    if auth_method == 'OIDC':
        required_configs.append('oidc_id_provider')
        required_configs.append('oidc_client_id')
        required_configs.append('oidc_scopes')
    for required_config in required_configs:
        if not config.has_option('waiverdb', required_config):
            raise click.ClickException(config_error.format(required_config))
    if not config.has_option('waiverdb', 'resultsdb_api_url'):
        raise click.ClickException(config_error.format('resultsdb_api_url'))


def check_response(resp, data, result_id=None):
    if result_id:
        msg = 'for result with id {0}'.format(result_id)
    else:
        msg = 'for result with subject {0} and testcase {1}'.format(json.dumps(data['subject']),
                                                                    data['testcase'])
    if not resp.ok:
        try:
            error_msg = resp.json()['message']
        except (ValueError, KeyError):
            error_msg = resp.text
        raise click.ClickException(
            'Failed to create waiver {0}:\n{1}'
            .format(msg, error_msg))
    click.echo('Created waiver {0} {1}'.format(
        resp.json()['id'], msg))


@click.command(context_settings={'help_option_names': ['-h', '--help']})
@click.option('--config-file', '-C', default='/etc/waiverdb/client.conf',
              type=click.Path(exists=True),
              help='Specify a config file to use')
@click.option('--result-id', '-r', multiple=True, type=int,
              help='Specify one or more results to be waived')
@click.option('--subject', '-s',
              help='Specify one subject for a result to waive')
@click.option('--testcase', '-t',
              help='Specify a testcase for the subject')
@click.option('--product-version', '-p',
              help='Specify one of PDC\'s product version identifiers.')
@click.option('--waived/--no-waived', default=True,
              help='Whether or not the result is waived')
@click.option('--comment', '-c',
              help='A comment explaining why the result is waived')
def cli(comment, waived, product_version, testcase, subject, result_id, config_file):
    """
    Creates new waiver against test results.

    Examples:

        waiverdb-cli -r 123 -r 456 -p "fedora-26" -c "It's dead!"

        or

        waiverdb-cli -t dist.rpmlint -s '{"item": "python-requests-1.2.3-1.fc26",
                                          "type": "koji_build"}'
                     -p "fedora-26" -c "It's dead!"

    """
    config = configparser.SafeConfigParser()

    config.read(config_file)
    validate_config(config)

    result_ids = result_id
    if not product_version:
        raise click.ClickException('Please specify product version')
    if result_ids and (subject or testcase):
        raise click.ClickException('Please specify result_id or subject/testcase. Not both')
    if not result_ids and not subject:
        raise click.ClickException('Please specify one subject')
    if not result_ids and not testcase:
        raise click.ClickException('Please specify testcase')

    auth_method = config.get('waiverdb', 'auth_method')
    data_list = []
    if not result_ids:
        data_list.append({
            'subject': json.loads(subject),
            'testcase': testcase,
            'waived': waived,
            'product_version': product_version,
            'comment': comment
        })

    # XXX - TODO - remove this in a future release.  (for backwards compat)
    for result_id in result_ids:
        data_list.append({
            'result_id': result_id,
            'waived': waived,
            'product_version': product_version,
            'comment': comment
        })

    api_url = config.get('waiverdb', 'api_url')
    if auth_method == 'OIDC':
        # Try to import this now so the user gets immediate feedback if
        # it isn't installed
        try:
            import openidc_client  # noqa: F401
        except ImportError:
            raise click.ClickException('python-openidc-client needs to be installed')
        # Get the auth token using the OpenID client.
        oidc_client_secret = None
        if config.has_option('waiverdb', 'oidc_client_secret'):
            oidc_client_secret = config.get('waiverdb', 'oidc_client_secret')
        oidc = openidc_client.OpenIDCClient(
            'waiverdb',
            config.get('waiverdb', 'oidc_id_provider'),
            {'Token': 'Token', 'Authorization': 'Authorization'},
            config.get('waiverdb', 'oidc_client_id'),
            oidc_client_secret)
        scopes = config.get('waiverdb', 'oidc_scopes').strip().splitlines()

        for data in data_list:
            resp = oidc.send_request(
                scopes=scopes,
                url='{0}/waivers/'.format(api_url.rstrip('/')),
                data=json.dumps(data),
                headers={'Content-Type': 'application/json'},
                timeout=60)
            check_response(resp, data, data.get('result_id', None))
    elif auth_method == 'Kerberos':
        # Try to import this now so the user gets immediate feedback if
        # it isn't installed
        try:
            import requests_kerberos  # noqa: F401
        except ImportError:
            raise click.ClickException('python-requests-kerberos needs to be installed')
        auth = requests_kerberos.HTTPKerberosAuth(mutual_authentication=requests_kerberos.OPTIONAL)
        for data in data_list:
            resp = requests.request('POST', '{0}/waivers/'.format(api_url.rstrip('/')),
                                    data=json.dumps(data), auth=auth,
                                    headers={'Content-Type': 'application/json'},
                                    timeout=60)
            if resp.status_code == 401:
                raise click.ClickException('WaiverDB authentication using Kerberos failed. '
                                           'Make sure you have a valid Kerberos ticket.')
            check_response(resp, data, data.get('result_id', None))
    elif auth_method == 'dummy':
        for data in data_list:
            resp = requests.request('POST', '{0}/waivers/'.format(api_url.rstrip('/')),
                                    data=json.dumps(data), auth=('user', 'pass'),
                                    headers={'Content-Type': 'application/json'},
                                    timeout=60)
            check_response(resp, data, data.get('result_id', None))


if __name__ == '__main__':
    cli()  # pylint: disable=E1120



# SPDX-License-Identifier: GPL-2.0+

"""This module contains tests for :mod:`waiverdb.events`."""
from __future__ import unicode_literals
import mock
from waiverdb.models import Waiver


@mock.patch('waiverdb.events.fedmsg')
def test_publish_new_waiver_with_fedmsg(mock_fedmsg, session):
    waiver = Waiver(
        subject={'subject.test': 'subject'},
        testcase='testcase1',
        username='jcline',
        product_version='something',
        waived=True,
        comment='This is a comment',
    )
    sesh = session()
    sesh.add(waiver)
    sesh.commit()
    mock_fedmsg.publish.assert_called_once_with(
        topic='waiver.new',
        msg={
            'id': waiver.id,
            'subject': {'subject.test': 'subject'},
            'testcase': 'testcase1',
            'username': 'jcline',
            'proxied_by': None,
            'product_version': 'something',
            'waived': True,
            'comment': 'This is a comment',
            'timestamp': waiver.timestamp.isoformat(),
        }
    )


@mock.patch('waiverdb.events.fedmsg')
def test_publish_new_waiver_with_fedmsg_for_proxy_user(mock_fedmsg, session):
    waiver = Waiver(
        subject={'subject.test': 'subject'},
        testcase='testcase1',
        username='jcline',
        product_version='something',
        waived=True,
        comment='This is a comment',
        proxied_by='bodhi'
    )
    sesh = session()
    sesh.add(waiver)
    sesh.commit()
    mock_fedmsg.publish.assert_called_once_with(
        topic='waiver.new',
        msg={
            'id': waiver.id,
            'subject': {'subject.test': 'subject'},
            'testcase': 'testcase1',
            'username': 'jcline',
            'proxied_by': 'bodhi',
            'product_version': 'something',
            'waived': True,
            'comment': 'This is a comment',
            'timestamp': waiver.timestamp.isoformat(),
        }
    )



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: GPL-2.0+

import socket

hostname = socket.gethostname()

config = dict(
    active=True,
    # Set this to dev if you're hacking on fedmsg or an app.
    # Set to stg or prod if running in the Fedora Infrastructure
    environment="dev",

    # Default is 0
    high_water_mark=0,
    io_threads=1,

    ## For the fedmsg-hub and fedmsg-relay. ##

    # This is a status dir to keep a record of the last processed message
    #status_directory=os.getcwd() + "/status",
    #status_directory='/var/run/fedmsg/status',

    # This is the URL of a datagrepper instance that we can query for backlog.
    #datagrepper_url="https://apps.fedoraproject.org/datagrepper/raw",

    # We almost always want the fedmsg-hub to be sending messages with zmq as
    # opposed to amqp or stomp.  You can send with only *one* of the messaging
    # backends: zeromq or amqp or stomp.  You cannot send with two or more at
    # the same time.  Here, zmq is either enabled, or it is not.  If it is not,
    # see the options below for how to configure stomp or amqp.
    zmq_enabled=True,

    # On the other hand, if you wanted to use STOMP *instead* of zeromq, you
    # could do the following...
    #zmq_enabled=False,
    #stomp_uri='localhost:59597,localhost:59598',
    #stomp_user='username',
    #stomp_pass='password',
    #stomp_ssl_crt='/path/to/an/optional.crt',
    #stomp_ssl_key='/path/to/an/optional.key',

    # When subscribing to messages, we want to allow splats ('*') so we tell
    # the hub to not be strict when comparing messages topics to subscription
    # topics.
    zmq_strict=False,

    # Number of seconds to sleep after initializing waiting for sockets to sync
    post_init_sleep=0.5,

    # Wait a whole second to kill all the last io threads for messages to
    # exit our outgoing queue (if we have any).  This is in milliseconds.
    zmq_linger=1000,

    # See the following
    #   - http://tldp.org/HOWTO/TCP-Keepalive-HOWTO/overview.html
    #   - http://api.zeromq.org/3-2:zmq-setsockopt
    zmq_tcp_keepalive=1,
    zmq_tcp_keepalive_cnt=3,
    zmq_tcp_keepalive_idle=60,
    zmq_tcp_keepalive_intvl=5,

    # Number of miliseconds that zeromq will wait to reconnect until it gets
    # a connection if an endpoint is unavailable.
    zmq_reconnect_ivl=100,
    # Max delay that you can reconfigure to reduce reconnect storm spam. This
    # is in miliseconds.
    zmq_reconnect_ivl_max=1000,

    # This is a dict of possible addresses from which fedmsg can send
    # messages.  fedmsg.init(...) requires that a 'name' argument be passed
    # to it which corresponds with one of the keys in this dict.
    endpoints={
        "waiverdb.%s" % hostname: [
            "tcp://127.0.0.1:5011",
        ],
        "relay_outbound": [
            "tcp://127.0.0.1:4001",
        ],
    },
    # This is the address of an active->passive relay.  It is used for the
    # fedmsg-logger command which requires another service with a stable
    # listening address for it to send messages to.
    # It is also used by the git-hook, for the same reason.
    # It is also used by the mediawiki php plugin which, due to the oddities of
    # php, can't maintain a single passive-bind endpoint of it's own.
    relay_inbound=[
        "tcp://127.0.0.1:2003",
    ],
    sign_messages=False,
    validate_signatures=False,

    # Use these implementations to sign and validate messages
    crypto_backend='x509',
    crypto_validate_backends=['x509'],

    ssldir="/etc/pki/fedmsg",
    crl_location="https://fedoraproject.org/fedmsg/crl.pem",
    crl_cache="/var/run/fedmsg/crl.pem",
    crl_cache_expiry=10,

    ca_cert_location="https://fedoraproject.org/fedmsg/ca.crt",
    ca_cert_cache="/var/run/fedmsg/ca.crt",
    ca_cert_cache_expiry=0,  # Never expires

    certnames={
        # In prod/stg, map hostname to the name of the cert in ssldir.
        # Unfortunately, we can't use socket.getfqdn()
        #"app01.stg": "app01.stg.phx2.fedoraproject.org",
    },

    # A mapping of fully qualified topics to a list of cert names for which
    # a valid signature is to be considered authorized.  Messages on topics not
    # listed here are considered automatically authorized.
    routing_policy={
        # Only allow announcements from production if they're signed by a
        # certain certificate.
        "org.fedoraproject.prod.announce.announcement": [
            "announce-lockbox.phx2.fedoraproject.org",
        ],
    },

    # Set this to True if you want messages to be dropped that aren't
    # explicitly whitelisted in the routing_policy.
    # When this is False, only messages that have a topic in the routing_policy
    # but whose cert names aren't in the associated list are dropped; messages
    # whose topics do not appear in the routing_policy are not dropped.
    routing_nitpicky=False,
)



"""add proxyuser waiving support

Revision ID: 0a74cdab732a
Revises: 0a27a8ad723a
Create Date: 2017-11-22 13:31:50.213681

"""

# revision identifiers, used by Alembic.
revision = '0a74cdab732a'
down_revision = '0a27a8ad723a'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('waiver', sa.Column('proxied_by', sa.String(length=255), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('waiver', 'proxied_by')
    # ### end Alembic commands ###



# SPDX-License-Identifier: GPL-2.0+

import pytest
import kerberos
import mock
import json
from werkzeug.exceptions import Unauthorized
import waiverdb.auth
import flask_oidc


@pytest.mark.usefixtures('enable_kerberos')
class TestKerberosAuthentication(object):

    def test_keytab_file_is_not_set_should_raise_error(self):
        with pytest.raises(Unauthorized):
            request = mock.MagicMock()
            headers = {'Authorization': "babablaba"}
            request.headers.return_value = mock.MagicMock(spec_set=dict)
            request.headers.__getitem__.side_effect = headers.__getitem__
            request.headers.__setitem__.side_effect = headers.__setitem__
            request.headers.__contains__.side_effect = headers.__contains__
            waiverdb.auth.get_user(request)

    def test_unauthorized(self, client, monkeypatch):
        monkeypatch.setenv('KRB5_KTNAME', '/etc/foo.keytab')
        r = client.post('/api/v1.0/waivers/', content_type='application/json')
        assert r.status_code == 401
        assert r.headers.get('www-authenticate') == 'Negotiate'

    @mock.patch('kerberos.authGSSServerInit', return_value=(kerberos.AUTH_GSS_COMPLETE, object()))
    @mock.patch('kerberos.authGSSServerStep', return_value=kerberos.AUTH_GSS_COMPLETE)
    @mock.patch('kerberos.authGSSServerResponse', return_value='STOKEN')
    @mock.patch('kerberos.authGSSServerUserName', return_value='foo@EXAMPLE.ORG')
    @mock.patch('kerberos.authGSSServerClean')
    @mock.patch('kerberos.getServerPrincipalDetails')
    def test_authorized(self, principal, clean, name, response, step, init,
                        client, monkeypatch, session):
        monkeypatch.setenv('KRB5_KTNAME', '/etc/foo.keytab')
        data = {
            'subject': {'subject.test': 'subject'},
            'testcase': 'testcase1',
            'product_version': 'fool-1',
            'waived': True,
            'comment': 'it broke',
        }
        r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                        content_type='application/json',
                        headers={'Authorization': 'Negotiate CTOKEN'})
        assert r.status_code == 201
        assert r.headers.get('WWW-Authenticate') == 'negotiate STOKEN'
        res_data = json.loads(r.data.decode('utf-8'))
        assert res_data['username'] == 'foo'


class TestOIDCAuthentication(object):

    def test_get_user_without_token(self, session):
        with pytest.raises(Unauthorized) as excinfo:
            request = mock.MagicMock()
            waiverdb.auth.get_user(request)
        assert "No 'Authorization' header found" in excinfo.value.get_description()

    @mock.patch.object(flask_oidc.OpenIDConnect, '_get_token_info')
    def test_get_user_with_invalid_token(self, mocked_get_token, session):
        # http://vsbattles.wikia.com/wiki/Son_Goku
        name = 'Son Goku'
        mocked_get_token.return_value = {'active': False, 'username': name,
                                         'scope': 'openid waiverdb_scope'}
        headers = {'Authorization': 'Bearer invalid'}
        request = mock.MagicMock()
        request.headers.return_value = mock.MagicMock(spec_set=dict)
        request.headers.__getitem__.side_effect = headers.__getitem__
        request.headers.__setitem__.side_effect = headers.__setitem__
        request.headers.__contains__.side_effect = headers.__contains__
        with pytest.raises(Unauthorized) as excinfo:
            waiverdb.auth.get_user(request)
        assert 'Token required but invalid' in excinfo.value.get_description()

    @mock.patch.object(flask_oidc.OpenIDConnect, '_get_token_info')
    def test_get_user_good(self, mocked_get_token, session):
        # http://vsbattles.wikia.com/wiki/Son_Goku
        name = 'Son Goku'
        mocked_get_token.return_value = {'active': True, 'username': name,
                                         'scope': 'openid waiverdb_scope'}
        headers = {'Authorization': 'Bearer foobar'}
        request = mock.MagicMock()
        request.headers.return_value = mock.MagicMock(spec_set=dict)
        request.headers.__getitem__.side_effect = headers.__getitem__
        request.headers.__setitem__.side_effect = headers.__setitem__
        request.headers.__contains__.side_effect = headers.__contains__
        user, header = waiverdb.auth.get_user(request)
        assert user == name


@pytest.mark.usefixtures('enable_ssl')
class TestSSLAuthentication(object):
    def test_SSL_CLIENT_VERIFY_is_not_set_should_raise_error(self):
        with pytest.raises(Unauthorized) as excinfo:
            request = mock.MagicMock()
            waiverdb.auth.get_user(request)
        assert 'Cannot verify client' in excinfo.value.get_description()

    def test_SSL_CLIENT_S_DN_is_not_set_should_raise_error(self):
        with pytest.raises(Unauthorized) as excinfo:
            request = mock.MagicMock(environ={'SSL_CLIENT_VERIFY': 'SUCCESS'})
            waiverdb.auth.get_user(request)
        assert 'Unable to get user information (DN) from the client certificate' \
               in excinfo.value.get_description()

    def test_good_ssl_cert(self):
        # http://vsbattles.wikia.com/wiki/Son_Goku
        name = 'Son Goku'
        ssl = {
            'SSL_CLIENT_VERIFY': 'SUCCESS',
            'SSL_CLIENT_S_DN': name
        }
        request = mock.MagicMock(environ=ssl)
        user, header = waiverdb.auth.get_user(request)
        assert user == name



"""Initializes migration

Revision ID: 0a27a8ad723a
Revises: None
Create Date: 2017-11-21 10:57:17.056587

"""

# revision identifiers, used by Alembic.
revision = '0a27a8ad723a'
down_revision = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('waiver',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('result_id', sa.Integer(), nullable=False),
    sa.Column('username', sa.String(length=255), nullable=False),
    sa.Column('product_version', sa.String(length=200), nullable=False),
    sa.Column('waived', sa.Boolean(), nullable=False),
    sa.Column('comment', sa.Text(), nullable=True),
    sa.Column('timestamp', sa.DateTime(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('waiver')
    # ### end Alembic commands ###



# SPDX-License-Identifier: GPL-2.0+

import logging
import sys
import systemd.journal


def log_to_stdout(app, level=logging.INFO):
    fmt = '[%(filename)s:%(lineno)d] ' if app.debug else '%(module)-12s '
    fmt += '%(asctime)s %(levelname)-7s %(message)s'
    datefmt = '%Y-%m-%d %H:%M:%S'
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setLevel(level)
    stream_handler.setFormatter(logging.Formatter(fmt=fmt, datefmt=datefmt))
    app.logger.addHandler(stream_handler)


def log_to_journal(app, level=logging.INFO):
    journal_handler = systemd.journal.JournalHandler()
    journal_handler.setLevel(level)
    app.logger.addHandler(journal_handler)


def init_logging(app):
    log_level = logging.DEBUG if app.debug else logging.INFO
    if app.config['JOURNAL_LOGGING']:
        log_to_journal(app, level=log_level)
    else:
        log_to_stdout(app, level=log_level)



# SPDX-License-Identifier: GPL-2.0+


import os
import kerberos
from flask import current_app, Response, g
# Starting with Flask 0.9, the _app_ctx_stack is the correct one,
# before that we need to use the _request_ctx_stack.
try:
    from flask import _app_ctx_stack as stack
except ImportError:
    from flask import _request_ctx_stack as stack
from socket import gethostname
from werkzeug.exceptions import Unauthorized, Forbidden


# Inspired by https://github.com/mkomitee/flask-kerberos/blob/master/flask_kerberos.py
class KerberosAuthenticate(object):

    def __init__(self):
        if current_app.config['KERBEROS_HTTP_HOST']:
            hostname = current_app.config['KERBEROS_HTTP_HOST']
        else:
            hostname = gethostname()
        self.service_name = "HTTP@%s" % (hostname)
        if 'KRB5_KTNAME' in os.environ:
            try:
                principal = kerberos.getServerPrincipalDetails('HTTP', hostname)
            except kerberos.KrbError as exc:
                raise Unauthorized("Authentication Kerberos Failure: %s" % exc.message[0])
            else:
                current_app.logger.debug("Kerberos: server is identifying as %s" % principal)
        else:
            raise Unauthorized("Kerberos: set KRB5_KTNAME to your keytab file")

    def _gssapi_authenticate(self, token):
        '''
        Performs GSSAPI Negotiate Authentication
        On success also stashes the server response token for mutual authentication
        at the top of request context with the name kerberos_token, along with the
        authenticated user principal with the name kerberos_user.
        '''
        state = None
        ctx = stack.top
        try:
            rc, state = kerberos.authGSSServerInit(self.service_name)
            if rc != kerberos.AUTH_GSS_COMPLETE:
                current_app.logger.error('Unable to initialize server context')
                return None
            rc = kerberos.authGSSServerStep(state, token)
            if rc == kerberos.AUTH_GSS_COMPLETE:
                current_app.logger.debug('Completed GSSAPI negotiation')
                ctx.kerberos_token = kerberos.authGSSServerResponse(state)
                ctx.kerberos_user = kerberos.authGSSServerUserName(state)
                return rc
            elif rc == kerberos.AUTH_GSS_CONTINUE:
                current_app.logger.debug('Continuing GSSAPI negotiation')
                return kerberos.AUTH_GSS_CONTINUE
            else:
                current_app.logger.debug('Unable to step server context')
                return None
        except kerberos.GSSError as e:
            current_app.logger.error('Unable to authenticate: %s', e)
            return None
        finally:
            if state:
                kerberos.authGSSServerClean(state)

    def process_request(self, token):
        """
        Authenticates the current request using Kerberos.
        """
        kerberos_user = None
        kerberos_token = None
        ctx = stack.top
        rc = self._gssapi_authenticate(token)
        if rc == kerberos.AUTH_GSS_COMPLETE:
            kerberos_user = ctx.kerberos_user
            kerberos_token = ctx.kerberos_token
        elif rc != kerberos.AUTH_GSS_CONTINUE:
            raise Forbidden("Invalid Kerberos ticket")
        return kerberos_user, kerberos_token


def get_user(request):
    user = None
    headers = dict()
    if current_app.config['AUTH_METHOD'] == 'OIDC':
        if 'Authorization' not in request.headers:
            raise Unauthorized("No 'Authorization' header found.")
        token = request.headers.get("Authorization").strip()
        prefix = 'Bearer '
        if not token.startswith(prefix):
            raise Unauthorized('Authorization headers must start with %r' % prefix)
        token = token[len(prefix):].strip()
        required_scopes = [
            'openid',
            current_app.config['OIDC_REQUIRED_SCOPE'],
        ]
        validity = current_app.oidc.validate_token(token, required_scopes)
        if validity is not True:
            raise Unauthorized(validity)
        user = g.oidc_token_info['username']
    elif current_app.config['AUTH_METHOD'] == 'Kerberos':
        if 'Authorization' not in request.headers:
            response = Response('Unauthorized', 401, {'WWW-Authenticate': 'Negotiate'})
            raise Unauthorized(response=response)
        header = request.headers.get("Authorization")
        token = ''.join(header.strip().split()[1:])
        user, kerberos_token = KerberosAuthenticate().process_request(token)
        # remove realm
        user = user.split("@")[0]
        if kerberos_token is not None:
            headers = {'WWW-Authenticate': ' '.join(['negotiate', kerberos_token])}
    elif current_app.config['AUTH_METHOD'] == 'SSL':
        # Nginx sets SSL_CLIENT_VERIFY and SSL_CLIENT_S_DN in request.environ
        # when doing SSL authentication.
        ssl_client_verify = request.environ.get('SSL_CLIENT_VERIFY')
        if ssl_client_verify != 'SUCCESS':
            raise Unauthorized('Cannot verify client: %s' % ssl_client_verify)
        if not request.environ.get('SSL_CLIENT_S_DN'):
            raise Unauthorized('Unable to get user information (DN) from the client certificate')
        user = request.environ.get('SSL_CLIENT_S_DN')
    elif current_app.config['AUTH_METHOD'] == 'dummy':
        # Blindly accept any username. For testing purposes only of course!
        if not request.authorization:
            response = Response('Unauthorized', 401, {'WWW-Authenticate': 'Basic realm="dummy"'})
            raise Unauthorized(response=response)
        user = request.authorization.username
    else:
        raise Unauthorized("Authenticated user required")
    return user, headers



# SPDX-License-Identifier: GPL-2.0+

import time
import click
from flask.cli import FlaskGroup
from sqlalchemy.exc import OperationalError
from waiverdb.models import db


def create_waiver_app(_):
    from waiverdb.app import create_app  # noqa: F401
    return create_app()


@click.group(cls=FlaskGroup, create_app=create_waiver_app)
def cli():
    pass


@cli.command(name='wait-for-db')
def wait_for_db():
    """
    Wait until database server is reachable.
    """
    poll_interval = 10  # seconds
    while True:
        try:
            db.engine.connect()
        except OperationalError as e:
            click.echo('Failed to connect to database: {}'.format(e))
            click.echo('Sleeping for {} seconds...'.format(poll_interval))
            time.sleep(poll_interval)
            click.echo('Retrying...')
        else:
            break


if __name__ == '__main__':
    cli()  # pylint: disable=E1120




# SPDX-License-Identifier: GPL-2.0+

from collections import OrderedDict
from waiverdb.models import Waiver


# https://pagure.io/waiverdb/issue/134
def test_waiver_subject_json_serialization_is_consistent(db):
    # With the default json.dumps serializer, the order of keys in the
    # serialized JSON object could vary because dict keys are iterated in
    # arbitrary order. Hence we have a custom serializer to guarantee the
    # serialization is predictable, regardless of iteration order.
    # In this test we therefore use OrderedDict and not dict, specifically so
    # we can guarantee that the iteration order is *not* the same in each
    # object.
    nvr = 'python-requests-1.2.3-1.fc26'
    subject = OrderedDict([('item', nvr), ('type', 'koji_build')])
    equivalent_subject = OrderedDict([('type', 'koji_build'), ('item', nvr)])
    waiver = Waiver(username='dcallagh', waived=True, comment='test', product_version='fedora-26',
                    testcase='dist.rpmlint', subject=subject)
    db.session.add(waiver)
    waiver = Waiver(username='dcallagh', waived=True, comment='test', product_version='fedora-26',
                    testcase='dist.rpmlint', subject=equivalent_subject)
    db.session.add(waiver)
    db.session.flush()
    query = Waiver.query.filter(Waiver.subject == {'item': nvr, 'type': 'koji_build'})
    # The query should select *both* the waivers above
    assert query.count() == 2



# SPDX-License-Identifier: GPL-2.0+

from flask_restful import fields


waiver_fields = {
    'id': fields.Integer,
    'subject': fields.Raw,
    'testcase': fields.String,
    'username': fields.String,
    'proxied_by': fields.String,
    'product_version': fields.String,
    'waived': fields.Boolean,
    'comment': fields.String,
    'timestamp': fields.DateTime(dt_format='iso8601'),
}



# SPDX-License-Identifier: GPL-2.0+
import pytest
import json
from mock import Mock, patch
from click.testing import CliRunner
from waiverdb.cli import cli as waiverdb_cli


def test_misconfigured_auth_method(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
api_url=http://localhost:5004/api/v1.0
        """)
    runner = CliRunner()
    args = ['-C', p.strpath]
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: The config option "auth_method" is required\n'


def test_misconfigured_api_url(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=OIDC
        """)
    runner = CliRunner()
    args = ['-C', p.strpath]
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: The config option "api_url" is required\n'


def test_misconfigured_resultdb_api_url(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=dummy
api_url=http://localhost:5004/api/v1.0
        """)
    runner = CliRunner()
    args = ['-C', p.strpath]
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: The config option "resultsdb_api_url" is required\n'


def test_misconfigured_oidc_id_provider(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=OIDC
api_url=http://localhost:5004/api/v1.0
        """)
    runner = CliRunner()
    args = ['-C', p.strpath]
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: The config option "oidc_id_provider" is required\n'


def test_misconfigured_oidc_client_id(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=OIDC
api_url=http://localhost:5004/api/v1.0
oidc_id_provider=https://id.stg.fedoraproject.org/openidc/
        """)
    runner = CliRunner()
    args = ['-C', p.strpath]
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: The config option "oidc_client_id" is required\n'


def test_misconfigured_oidc_scopes(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=OIDC
api_url=http://localhost:5004/api/v1.0
oidc_id_provider=https://id.stg.fedoraproject.org/openidc/
oidc_client_id=waiverdb
        """)
    runner = CliRunner()
    args = ['-C', p.strpath]
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: The config option "oidc_scopes" is required\n'


def test_no_product_version(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=OIDC
api_url=http://localhost:5004/api/v1.0
oidc_id_provider=https://id.stg.fedoraproject.org/openidc/
oidc_client_id=waiverdb
oidc_scopes=
    openid
resultsdb_api_url=http://localhost:5001/api/v2.0
        """)
    runner = CliRunner()
    args = ['-C', p.strpath]
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: Please specify product version\n'


def test_no_subject(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=OIDC
api_url=http://localhost:5004/api/v1.0
oidc_id_provider=https://id.stg.fedoraproject.org/openidc/
oidc_client_id=waiverdb
oidc_scopes=
    openid
resultsdb_api_url=http://localhost:5001/api/v2.0
        """)
    runner = CliRunner()
    args = ['-C', p.strpath, '-p', 'fedora-26']
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: Please specify one subject\n'


def test_no_testcase(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=OIDC
api_url=http://localhost:5004/api/v1.0
oidc_id_provider=https://id.stg.fedoraproject.org/openidc/
oidc_client_id=waiverdb
oidc_scopes=
    openid
resultsdb_api_url=http://localhost:5001/api/v2.0
        """)
    runner = CliRunner()
    args = ['-C', p.strpath, '-p', 'fedora-26', '-s', 'subject']
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: Please specify testcase\n'


def test_oidc_auth_is_enabled(tmpdir):
    # Skip if waiverdb is rebuilt for an environment where Kerberos authentication
    # is used and python-openidc-client is not available.
    pytest.importorskip('openidc_client')
    with patch('openidc_client.OpenIDCClient.send_request') as mock_oidc_req:
        mock_rv = Mock()
        mock_rv.json.return_value = {
            "comment": "It's dead!",
            "id": 15,
            "product_version": "Parrot",
            "subject": {"subject.test": "test", "s": "t"},
            "testcase": "test.testcase",
            "timestamp": "2017-010-16T17:42:04.209638",
            "username": "foo",
            "waived": True
        }
        mock_oidc_req.return_value = mock_rv
        p = tmpdir.join('client.conf')
        p.write("""
[waiverdb]
auth_method=OIDC
api_url=http://localhost:5004/api/v1.0
oidc_id_provider=https://id.stg.fedoraproject.org/openidc/
oidc_client_id=waiverdb
oidc_scopes=
    openid
resultsdb_api_url=http://localhost:5001/api/v2.0
            """)
        runner = CliRunner()
        args = ['-C', p.strpath, '-p', 'Parrot', '-s', '{"subject.test": "test", "s": "t"}',
                '-t', 'test.testcase', '-c', "It's dead!"]
        result = runner.invoke(waiverdb_cli, args)
        exp_json = {
            "subject": {"subject.test": "test", "s": "t"},
            "testcase": "test.testcase",
            'waived': True,
            'product_version': 'Parrot',
            'comment': "It's dead!"
        }
        mock_oidc_req.assert_called_once_with(
            url='http://localhost:5004/api/v1.0/waivers/',
            data=json.dumps(exp_json),
            scopes=['openid'],
            timeout=60,
            headers={'Content-Type': 'application/json'})
        assert result.exit_code == 0
        assert result.output == 'Created waiver 15 for result with subject {"subject.test": "test", "s": "t"} and testcase test.testcase\n' # noqa


def test_kerberos_is_enabled(tmpdir):
    # Skip if waiverdb is rebuilt for an environment where OIDC authentication
    # is used and python-requests-kerberos is not available.
    pytest.importorskip('requests_kerberos')
    with patch('requests.request') as mock_request:
        mock_rv = Mock()
        mock_rv.json.return_value = {
            "comment": "It's dead!",
            "id": 15,
            "product_version": "Parrot",
            "subject": {"subject.test": "test", "s": "t"},
            "testcase": "test.testcase",
            "timestamp": "2017-010-16T17:42:04.209638",
            "username": "foo",
            "waived": True
        }
        mock_request.return_value = mock_rv
        p = tmpdir.join('client.conf')
        p.write("""
[waiverdb]
auth_method=Kerberos
api_url=http://localhost:5004/api/v1.0
resultsdb_api_url=http://localhost:5001/api/v2.0
            """)
        runner = CliRunner()
        args = ['-C', p.strpath, '-p', 'Parrot', '-s', '{"subject.test": "test", "s": "t"}',
                '-t', 'test.testcase', '-c', "It's dead!"]
        result = runner.invoke(waiverdb_cli, args)
        mock_request.assert_called_once()
        assert result.output == 'Created waiver 15 for result with subject {"subject.test": "test", "s": "t"} and testcase test.testcase\n' # noqa


def test_submit_waiver_with_id(tmpdir):
    with patch('requests.request') as mock_request:
        mock_rv = Mock()
        mock_rv.json.return_value = {
            "comment": "It's dead!",
            "data": {"item": ["htop-1.0-1.fc22"], "type": ["bodhi_update"]},
            "id": 15,
            "product_version": "Parrot",
            "testcase": {"name": "test.testcase"},
            "timestamp": "2017-010-16T17:42:04.209638",
            "username": "foo",
            "waived": True
        }
        mock_request.return_value = mock_rv
        p = tmpdir.join('client.conf')
        p.write("""
[waiverdb]
auth_method=dummy
api_url=http://localhost:5004/api/v1.0
resultsdb_api_url=http://localhost:5001/api/v2.0
        """)
        runner = CliRunner()
        args = ['-C', p.strpath, '-p', 'Parrot', '-r', '123',
                '-c', "It's dead!"]
        result = runner.invoke(waiverdb_cli, args)
        mock_request.assert_called()
        assert result.output == 'Created waiver 15 for result with id 123\n'


def test_submit_waiver_with_multiple_ids(tmpdir):
    with patch('requests.request') as mock_request:
        mock_rv = Mock()
        mock_rv.json.return_value = {
            "comment": "It's dead!",
            "data": {"item": ["htop-1.0-1.fc22"], "type": ["bodhi_update"]},
            "id": 15,
            "testcase": {"name": "test.testcase"},
            "timestamp": "2017-010-16T17:42:04.209638",
            "username": "foo",
            "waived": True
        }
        mock_request.return_value = mock_rv
        p = tmpdir.join('client.conf')
        p.write("""
[waiverdb]
auth_method=dummy
api_url=http://localhost:5004/api/v1.0
resultsdb_api_url=http://localhost:5001/api/v2.0
        """)
        runner = CliRunner()
        args = ['-C', p.strpath, '-p', 'Parrot', '-r', '123', '-r', '456',
                '-c', "It's dead!"]
        result = runner.invoke(waiverdb_cli, args)
        mock_request.assert_called()

        assert result.output == 'Created waiver 15 for result with id 123\n\
Created waiver 15 for result with id 456\n'


def test_malformed_submission_with_id_and_subject_and_testcase(tmpdir):
        runner = CliRunner()
        p = tmpdir.join('client.conf')
        p.write("""
[waiverdb]
auth_method=dummy
api_url=http://localhost:5004/api/v1.0
resultsdb_api_url=http://localhost:5001/api/v2.0
        """)
        args = ['-C', p.strpath, '-p', 'Parrot', '-r', '123', '-s',
                '{"subject.test": "test", "s": "t"}', '-c', "It's dead!"]
        result = runner.invoke(waiverdb_cli, args)
        assert result.output == 'Error: Please specify result_id or subject/testcase. Not both\n'


def test_submit_waiver_for_original_spec_nvr_result(tmpdir):
    with patch('requests.request') as mock_request:
        mock_rv = Mock()
        mock_rv.json.return_value = {
            "comment": "It's dead!",
            "original_spec_nvr": "test",
            "id": 15,
            "product_version": "Parrot",
            "testcase": {"name": "test.testcase"},
            "timestamp": "2017-010-16T17:42:04.209638",
            "username": "foo",
            "waived": True
        }
        mock_request.return_value = mock_rv
        p = tmpdir.join('client.conf')
        p.write("""
[waiverdb]
auth_method=dummy
api_url=http://localhost:5004/api/v1.0
resultsdb_api_url=http://localhost:5001/api/v2.0
            """)
        runner = CliRunner()
        args = ['-C', p.strpath, '-p', 'Parrot', '-r', '123',
                '-c', "It's dead!"]
        result = runner.invoke(waiverdb_cli, args)
        mock_request.assert_called()
        assert result.output == 'Created waiver 15 for result with id 123\n'



# SPDX-License-Identifier: GPL-2.0+

import datetime
from .base import db, EqualityComparableJSONType
from sqlalchemy import or_, and_, cast


class Waiver(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    result_id = db.Column(db.Integer, nullable=True)
    subject = db.Column(EqualityComparableJSONType, nullable=False)
    testcase = db.Column(db.Text, nullable=False, index=True)
    username = db.Column(db.String(255), nullable=False)
    proxied_by = db.Column(db.String(255))
    product_version = db.Column(db.String(200), nullable=False)
    waived = db.Column(db.Boolean, nullable=False, default=False)
    comment = db.Column(db.Text)
    timestamp = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    __table_args__ = (
        db.Index('ix_waiver_subject', cast(subject, db.Text)),
    )

    def __init__(self, subject, testcase, username, product_version, waived=False,
                 comment=None, proxied_by=None):
        self.subject = subject
        self.testcase = testcase
        self.username = username
        self.product_version = product_version
        self.waived = waived
        self.comment = comment
        self.proxied_by = proxied_by

    def __repr__(self):
        return '%s(result_id=%r, subject=%r, testcase=%r, username=%r, product_version=%r,\
                waived=%r)' % (self.__class__.__name__, self.result_id, self.subject, self.testcase,
                               self.username, self.product_version, self.waived)

    @classmethod
    def by_results(cls, query, results):
        return query.filter(or_(*[
            and_(
                cls.subject == d['subject'],
                cls.testcase == d['testcase']
            ) if d.get('testcase', None) else
            and_(
                cls.subject == d['subject']
            ) if d.get('subject', None) else
            and_() for d in results
        ]))



# SPDX-License-Identifier: GPL-2.0+

import os
import re
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md')) as fd:
    README = fd.read()


def get_project_version(version_file='waiverdb/__init__.py'):
    """
    Read the declared version of the project from the source code.

    Args:
        version_file: The file with the version string in it. The version must
            be in the format ``__version__ = '<version>'`` and the file must be
            UTF-8 encoded.
    """
    with open(version_file, 'r') as f:
        version_pattern = "^__version__ = '(.+)'$"
        match = re.search(version_pattern, f.read(), re.MULTILINE)
    if match is None:
        err_msg = 'No line matching %r found in %r'
        raise ValueError(err_msg % (version_pattern, version_file))
    return match.group(1)


setup(name='waiverdb',
      version=get_project_version(),
      description='An engine for storing waivers against test results.',
      long_description=README,
      author='Red Hat, Inc.',
      author_email='qa-devel@lists.fedoraproject.org',
      license='GPLv2+',
      packages=find_packages(exclude=['tests']),
      include_package_data=True,
      entry_points={
          'console_scripts': ['waiverdb-cli=waiverdb.cli:cli',
                              'waiverdb=waiverdb.manage:cli'],
      },
)



# SPDX-License-Identifier: GPL-2.0+

import json

import requests
from flask import Blueprint, request, current_app
from flask_restful import Resource, Api, reqparse, marshal_with, marshal
from werkzeug.exceptions import BadRequest, UnsupportedMediaType, Forbidden, ServiceUnavailable
from sqlalchemy.sql.expression import func, cast

from waiverdb import __version__
from waiverdb.models import db, Waiver
from waiverdb.utils import reqparse_since, json_collection, jsonp
from waiverdb.fields import waiver_fields
import waiverdb.auth

api_v1 = (Blueprint('api_v1', __name__))
api = Api(api_v1)
requests_session = requests.Session()


def valid_dict(value):
    if not isinstance(value, dict):
        raise ValueError("Must be a valid dict, not %r" % value)
    return value


def get_resultsdb_result(result_id):
    response = requests_session.request('GET', '{0}/results/{1}'.format(
        current_app.config['RESULTSDB_API_URL'], result_id),
        headers={'Content-Type': 'application/json'},
        timeout=60)
    response.raise_for_status()
    return response.json()


# RP contains request parsers (reqparse.RequestParser).
#    Parsers are added in each 'resource section' for better readability
RP = {}
RP['create_waiver'] = reqparse.RequestParser()
RP['create_waiver'].add_argument('subject', type=valid_dict, location='json')
RP['create_waiver'].add_argument('testcase', type=str, location='json')
RP['create_waiver'].add_argument('result_id', type=int, location='json')
RP['create_waiver'].add_argument('waived', type=bool, required=True, location='json')
RP['create_waiver'].add_argument('product_version', type=str, required=True, location='json')
RP['create_waiver'].add_argument('comment', type=str, default=None, location='json')
RP['create_waiver'].add_argument('username', type=str, default=None, location='json')

RP['get_waivers'] = reqparse.RequestParser()
RP['get_waivers'].add_argument('results', location='args')
RP['get_waivers'].add_argument('product_version', location='args')
RP['get_waivers'].add_argument('username', location='args')
RP['get_waivers'].add_argument('include_obsolete', type=bool, default=False, location='args')
# XXX This matches the since query parameter in resultsdb but I think it would
# be good to use two parameters(since and until).
RP['get_waivers'].add_argument('since', location='args')
RP['get_waivers'].add_argument('page', default=1, type=int, location='args')
RP['get_waivers'].add_argument('limit', default=10, type=int, location='args')
RP['get_waivers'].add_argument('proxied_by', location='args')


class WaiversResource(Resource):
    @jsonp
    def get(self):
        """
        Get waiver records.

        **Sample request**:

        .. sourcecode:: http

           GET /api/v1.0/waivers/ HTTP/1.1
           Host: localhost:5004
           User-Agent: curl/7.51.0
           Accept: application/json

        **Sample response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           Content-Length: 184
           Server: Werkzeug/0.12.1 Python/2.7.13
           Date: Thu, 16 Mar 2017 13:53:14 GMT

           {
               "data": [],
               "first": "http://localhost:5004/api/v1.0/waivers/?page=1",
               "last": "http://localhost:5004/api/v1.0/waivers/?page=0",
               "next": null,
               "prev": null
           }


        :query int page: The page to get.
        :query int limit: Limit the number of items returned.
        :query string results: Filter the waivers by result. Accepts a list of
            dictionaries, with one key 'subject' and one key 'testcase'.
        :query string product_version: Filter the waivers by product version.
        :query string username: Filter the waivers by username.
        :query string proxied_by: Filter the waivers by the users who are
            allowed to create waivers on behalf of other users.
        :query string since: An ISO 8601 formatted datetime (e.g. 2017-03-16T13:40:05+00:00)
            to filter results by. Optionally provide a second ISO 8601 datetime separated
            by a comma to retrieve a range (e.g. 2017-03-16T13:40:05+00:00,
            2017-03-16T13:40:15+00:00)
        :query boolean include_obsolete: If true, obsolete waivers will be included.
        :statuscode 200: If the query was valid and no problems were encountered.
            Note that the response may still contain 0 waivers.
        :statuscode 400: The request was malformed and could not be processed.
        """
        args = RP['get_waivers'].parse_args()
        query = Waiver.query.order_by(Waiver.timestamp.desc())

        if args['results']:
            results = json.loads(args['results'])
            for d in results:
                if d.get('subject', None):
                    if not isinstance(d.get('subject', None), dict):
                        raise BadRequest("'results' parameter should be a list \
                                          of dictionaries with subject and testcase")
                if d.get('testcase', None):
                    if not isinstance(d.get('testcase', None), basestring):
                        raise BadRequest("'results' parameter should be a list \
                                      of dictionaries with subject and testcase")
            query = Waiver.by_results(query, results)
        if args['product_version']:
            query = query.filter(Waiver.product_version == args['product_version'])
        if args['username']:
            query = query.filter(Waiver.username == args['username'])
        if args['proxied_by']:
            query = query.filter(Waiver.proxied_by == args['proxied_by'])
        if args['since']:
            try:
                since_start, since_end = reqparse_since(args['since'])
            except ValueError:
                raise BadRequest("'since' parameter not in ISO8601 format")
            except TypeError:
                raise BadRequest("'since' parameter not in ISO8601 format")
            if since_start:
                query = query.filter(Waiver.timestamp >= since_start)
            if since_end:
                query = query.filter(Waiver.timestamp <= since_end)
        if not args['include_obsolete']:
            subquery = db.session.query(func.max(Waiver.id)).group_by(cast(Waiver.subject, db.Text),
                                                                      Waiver.testcase)
            query = query.filter(Waiver.id.in_(subquery))
        query = query.order_by(Waiver.timestamp.desc())
        return json_collection(query, args['page'], args['limit'])

    @jsonp
    @marshal_with(waiver_fields)
    def post(self):
        """
        Create a new waiver.

        **Sample request**:

        .. sourcecode:: http

           POST /api/v1.0/waivers/ HTTP/1.1
           Host: localhost:5004
           Accept-Encoding: gzip, deflate
           Accept: application/json
           Connection: keep-alive
           User-Agent: HTTPie/0.9.4
           Content-Type: application/json
           Content-Length: 91

           {
               "subject": {"productmd.compose.id": "Fedora-9000-19700101.n.18"},
               "testcase": "compose.install_no_user",
               "waived": false,
               "product_version": "Parrot",
               "comment": "It's dead!"
           }



        **Sample response**:

        .. sourcecode:: http

           HTTP/1.0 201 CREATED
           Content-Length: 228
           Content-Type: application/json
           Date: Thu, 16 Mar 2017 17:42:04 GMT
           Server: Werkzeug/0.12.1 Python/2.7.13

           {
               "comment": "It's dead!",
               "id": 15,
               "product_version": "Parrot",
               "subject": {"productmd.compose.id": "Fedora-9000-19700101.n.18"},
               "testcase": "compose.install_no_user",
               "timestamp": "2017-03-16T17:42:04.209638",
               "username": "jcline",
               "waived": false,
               "proxied_by": null
           }

        :json object subject: The result subject for the waiver.
        :json string testcase: The result testcase for the waiver.
        :json boolean waived: Whether or not the result is waived.
        :json string product_version: The product version string.
        :json string comment: A comment explaining the waiver.
        :json string username: Username on whose behalf the caller is proxying.
        :statuscode 201: The waiver was successfully created.
        """

        user, headers = waiverdb.auth.get_user(request)
        args = RP['create_waiver'].parse_args()
        proxied_by = None
        if args.get('username'):
            if user not in current_app.config['SUPERUSERS']:
                raise Forbidden('user %s does not have the proxyuser ability' % user)
            proxied_by = user
            user = args['username']

        # XXX - remove this in a future release (it was for temp backwards compat)
        if args['result_id']:
            if args['subject'] or args['testcase']:
                raise BadRequest('Only result_id or subject and '
                                 'testcase are allowed.  Not both.')
            try:
                result = get_resultsdb_result(args['result_id'])
            except requests.HTTPError as e:
                if e.response.status_code == 404:
                    raise BadRequest('Result id not found in Resultsdb')
                else:
                    raise ServiceUnavailable('Failed looking up result in Resultsdb: %s' % e)
            except Exception as e:
                raise ServiceUnavailable('Failed looking up result in Resultsdb: %s' % e)
            if 'original_spec_nvr' in result['data']:
                subject = {'original_spec_nvr': result['data']['original_spec_nvr'][0]}
            else:
                if result['data']['type'][0] == 'koji_build' or \
                   result['data']['type'][0] == 'bodhi_update':
                    SUBJECT_KEYS = ['item', 'type']
                    subject = dict([(k, v[0]) for k, v in result['data'].items()
                                    if k in SUBJECT_KEYS])
                else:
                    raise BadRequest('It is not possible to submit a waiver by '
                                     'id for this result. Please try again specifying '
                                     'a subject and a testcase.')
            args['subject'] = subject
            args['testcase'] = result['testcase']['name']

        if not args['subject'] or not args['testcase']:
            raise BadRequest('Either result_id or subject/testcase '
                             'are required arguments.')

        waiver = Waiver(args['subject'], args['testcase'], user,
                        args['product_version'], args['waived'], args['comment'], proxied_by)
        db.session.add(waiver)
        db.session.commit()
        return waiver, 201, headers


class WaiverResource(Resource):
    @jsonp
    @marshal_with(waiver_fields)
    def get(self, waiver_id):
        """
        Get a single waiver by waiver ID.

        :param int waiver_id: The waiver's database ID.

        :statuscode 200: The waiver was found and returned.
        :statuscode 404: No waiver exists with that ID.
        """
        try:
            return Waiver.query.get_or_404(waiver_id)
        except Exception as NotFound:
            raise type(NotFound)('Waiver not found')


class GetWaiversBySubjectsAndTestcases(Resource):
    @jsonp
    def post(self):
        """
        Return a list of waivers by filtering the waivers with a
        list of result subjects and testcases.
        This accepts POST requests in order to handle a special
        case where a GET /waivers/ request has a long query string with many
        result subjects/testcases that could cause 413 errors.

        **Sample request**:

        .. sourcecode:: http

           POST /api/v1.0/waivers/+by-subjects-and-testcases HTTP/1.1
           Host: localhost:5004
           Accept-Encoding: gzip, deflate
           Accept: application/json
           Connection: keep-alive
           User-Agent: HTTPie/0.9.4
           Content-Type: application/json
           Content-Length: 40

           {
                "results": [
                    {
                        "subject": {"productmd.compose.id": "Fedora-9000-19700101.n.18"},
                        "testcase": "compose.install_no_user"
                    },
                    {
                        "subject": {"item": "gzip-1.9-1.fc28", "type": "koji_build"},
                        "testcase": "dist.rpmlint"
                    }
                ]
           }

        **Sample response**:

        .. sourcecode:: http

            HTTP/1.0 200 OK
            Content-Length: 562
            Content-Type: application/json
            Date: Thu, 21 Sep 2017 04:58:37 GMT
            Server: Werkzeug/0.11.10 Python/2.7.13

            {
                "data": [
                    {
                        "comment": "It's dead!",
                        "id": 5,
                        "product_version": "Parrot",
                        "subject": {"productmd.compose.id": "Fedora-9000-19700101.n.18"},
                        "testcase": "compose.install_no_user",
                        "timestamp": "2017-09-21T04:55:53.343368",
                        "username": "dummy",
                        "waived": true,
                        "proxied_by": null
                    },
                    {
                        "comment": "It's dead!",
                        "id": 4,
                        "product_version": "Parrot",
                        "subject": {"item": "gzip-1.9-1.fc28", "type": "koji_build"},
                        "testcase": "dist.rpmlint",
                        "timestamp": "2017-09-21T04:55:51.936658",
                        "username": "dummy",
                        "waived": true,
                        "proxied_by": null
                    }
                ]
            }

        :jsonparam array results: Filter the waivers by a list of dictionaries
            with result subjects and testcase.
        :jsonparam string product_version: Filter the waivers by product version.
        :jsonparam string username: Filter the waivers by username.
        :jsonparam string proxied_by: Filter the waivers by the users who are
            allowed to create waivers on behalf of other users.
        :jsonparam string since: An ISO 8601 formatted datetime (e.g. 2017-03-16T13:40:05+00:00)
            to filter results by. Optionally provide a second ISO 8601 datetime separated
            by a comma to retrieve a range (e.g. 2017-03-16T13:40:05+00:00,
            2017-03-16T13:40:15+00:00)
        :jsonparam boolean include_obsolete: If true, obsolete waivers will be included.
        :statuscode 200: If the query was valid and no problems were encountered.
            Note that the response may still contain 0 waivers.
        """
        if not request.get_json():
            raise UnsupportedMediaType('No JSON payload in request')
        data = request.get_json()
        query = Waiver.query.order_by(Waiver.timestamp.desc())
        if data.get('results'):
            for d in data['results']:
                if d.get('subject', None):
                    if not isinstance(d.get('subject', None), dict):
                        raise BadRequest("'results' parameter should be a list \
                                          of dictionaries with subject and testcase")
                if d.get('testcase', None):
                    if not isinstance(d.get('testcase', None), basestring):
                        raise BadRequest("'results' parameter should be a list \
                                          of dictionaries with subject and testcase")
            query = Waiver.by_results(query, data['results'])
        if 'product_version' in data:
            query = query.filter(Waiver.product_version == data['product_version'])
        if 'username' in data:
            query = query.filter(Waiver.username == data['username'])
        if 'proxied_by' in data:
            query = query.filter(Waiver.proxied_by == data['proxied_by'])
        if 'since' in data:
            try:
                since_start, since_end = reqparse_since(data['since'])
            except ValueError:
                raise BadRequest("'since' parameter not in ISO8601 format")
            except TypeError:
                raise BadRequest("'since' parameter not in ISO8601 format")
            if since_start:
                query = query.filter(Waiver.timestamp >= since_start)
            if since_end:
                query = query.filter(Waiver.timestamp <= since_end)
        if not data.get('include_obsolete', False):
            subquery = db.session.query(func.max(Waiver.id)).group_by(cast(Waiver.subject, db.Text),
                                                                      Waiver.testcase)
            query = query.filter(Waiver.id.in_(subquery))

        query = query.order_by(Waiver.timestamp.desc())
        return {'data': marshal(query.all(), waiver_fields)}


class AboutResource(Resource):
    @jsonp
    def get(self):
        """
        Returns the current running version and the method used for authentication.

        **Sample response**:

        .. sourcecode:: none

          HTTP/1.0 200 OK
          Content-Length: 55
          Content-Type: application/json
          Date: Tue, 31 Oct 2017 04:29:19 GMT
          Server: Werkzeug/0.11.10 Python/2.7.13

          {
            "auth_method": "OIDC",
            "version": "0.3.1"
          }

        :statuscode 200: Currently running waiverdb software version and authentication
                         are returned.
        """
        return {'version': __version__, 'auth_method': current_app.config['AUTH_METHOD']}


# set up the Api resource routing here
api.add_resource(WaiversResource, '/waivers/')
api.add_resource(WaiverResource, '/waivers/<int:waiver_id>')
api.add_resource(GetWaiversBySubjectsAndTestcases, '/waivers/+by-subjects-and-testcases')
api.add_resource(AboutResource, '/about')



# SPDX-License-Identifier: GPL-2.0+

import json
from sqlalchemy.dialects.postgresql import JSON
from sqlalchemy.dialects.postgresql.psycopg2 import _PGJSON
from sqlalchemy.sql.elements import ColumnElement, Null
from sqlalchemy.sql.expression import cast
from sqlalchemy.sql.sqltypes import Text
from flask_sqlalchemy import SQLAlchemy


json_serializer = json.encoder.JSONEncoder(sort_keys=True, separators=(',', ':')).encode


# Note that we have to inherit from the psycopg2-specific _PGJSON type,
# instead of the more general Postgres-specific JSON type. That's because
# SQLAlchemy helpfully "adapts" the general JSON type to _PGJSON when it sees
# we are using psycopg2, thereby defeating our customisations below.
# Inheriting from _PGJSON tells SQLAlchemy not do that adaptation.
# The magic occurs in SQLALchemy adapt_type() and PGDialect_psycopg2.colspecs.
class EqualityComparableJSONType(_PGJSON):
    """
    Like the standard JSON column type, but supports equality comparison of
    entire JSON structures. Postgres doesn't permit this natively on the JSON
    type so we have to instead cast to TEXT before comparing (and therefore we
    have to also go to lengths to ensure the JSON we feed in is consistently
    serialized).
    """

    class Comparator(JSON.Comparator):

        def __eq__(self, other):
            # If the RHS is a SQL expression, just assume its SQL type is
            # already JSON and thus CAST(AS TEXT) it as well.
            if isinstance(other, ColumnElement):
                rhs = cast(other, Text)
            else:
                rhs = json_serializer(other)
            return cast(self.expr, Text) == rhs

    comparator_factory = Comparator

    def bind_processor(self, dialect):
        def process(value):
            # JSON.NULL is a sentinel object meaning JSON "null" (SQLAlchemy 1.1+ only)
            if hasattr(self, 'NULL') and value is self.NULL:
                value = None
            elif isinstance(value, Null) or (value is None and self.none_as_null):
                return None
            return json_serializer(value)
        return process

    def result_processor(self, dialect, coltype):
        if dialect._has_native_json:  # pylint: disable=W0212
            return None

        def process(value):
            if value is None:
                return None
            return json.loads(value)
        return process


db = SQLAlchemy()
